from typing import Generator, Callable
from pyspark.sql import Column
import pyspark.sql.functions as fn
from pyspark.sql.types import BinaryType, IntegerType

import trend_etl.schema.schema as schema
from .udfs import double_to_binary


def binary_coalesce(col: Column) -> Column:
    return fn.coalesce(col, fn.lit("").cast(BinaryType()))


def array_to_binary(col: Column, to_binary: Callable[[Column], Column]) -> Column:
    return binary_coalesce(
        fn.aggregate(
            col,
            fn.lit("").cast(BinaryType()),
            lambda acc, elem: fn.concat(acc, binary_coalesce(to_binary(elem))),
        )
    )


def direct_to_binary(col: Column) -> Column:
    return binary_coalesce(col.cast(BinaryType()))


def boolean_to_binary(col: Column) -> Column:
    return direct_to_binary(col.cast(IntegerType()))


def discord_user_to_binary(user: Column) -> Column:
    return fn.concat(
        direct_to_binary(user.getField("id")),
        direct_to_binary(user.getField("username")),
        direct_to_binary(user.getField("global_name")),
        direct_to_binary(user.getField("avatar")),
        boolean_to_binary(user.getField("bot")),
        boolean_to_binary(user.getField("system")),
        direct_to_binary(user.getField("locale")),
        direct_to_binary(user.getField("email")),
        direct_to_binary(user.getField("flags")),
    )


def canonicalize_facebook_fields(
    data_after_norm: Column, field_specs: list[list[str]]
) -> Generator[Column, None, None]:
    for field_name, field_type in field_specs:
        match (field_name, field_type):
            case (_, "bigint" | "int" | "string" | "long"):
                yield direct_to_binary(data_after_norm.getField(field_name))
            # reactions
            case (_, "struct<summary struct<total_count long>>"):
                yield direct_to_binary(
                    data_after_norm.getField(field_name)
                    .getField("summary")
                    .getField("total_count")
                )
            case ("parent", "struct<id string>"):
                yield direct_to_binary(
                    data_after_norm.getField(field_name).getField("id")
                )
            case _:
                assert False, f"missed field {field_name}"


def canonicalize_reddit_fields(
    data_after_norm: Column, field_specs: list[list[str]]
) -> Generator[Column, None, None]:
    for field_name, field_type in field_specs:
        match (field_name, field_type):
            case (_, "boolean"):
                yield boolean_to_binary(data_after_norm.getField(field_name))
            case (_, "bigint" | "int" | "string" | "long"):
                yield direct_to_binary(data_after_norm.getField(field_name))
            case (_, "double"):
                yield double_to_binary(data_after_norm.getField(field_name))
            case (_, "array<string>" | "array<bigint>"):
                yield array_to_binary(
                    data_after_norm.getField(field_name), direct_to_binary
                )
            case ("gildings", "struct<gid_1:bigint,gid_2:bigint,gid_3:bigint>"):
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("gid_1")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("gid_2")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("gid_3")
                    ),
                )
            case (
                "author_flair_richtext",
                "array<struct<a:string,e:string,t:string,u:string>>",
            ):
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("a")),
                        direct_to_binary(elem.getField("e")),
                        direct_to_binary(elem.getField("t")),
                        direct_to_binary(elem.getField("u")),
                    ),
                )
            #
            # Subreddit cases
            #
            case (
                "comment_contribution_settings",
                "struct<allowed_media_types array<string>>",
            ):
                yield array_to_binary(
                    data_after_norm.getField(field_name).getField(
                        "allowed_media_types"
                    ),
                    direct_to_binary,
                )
            #
            # Submission cases
            #
            case ("link_flair_richtext", "array<struct<e:string,t:string>>"):
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("e")),
                        direct_to_binary(elem.getField("t")),
                    ),
                )
            #
            # Reply cases
            #
            case (
                "all_awardings",
                "array<struct<award_sub_type string,award_type string,awardings_required_to_grant_benefits bigint,coin_price bigint,coin_reward bigint,count bigint,days_of_drip_extension bigint,days_of_premium bigint,description string,end_date string,giver_coin_reward string,icon_format string,icon_height bigint,icon_url string,icon_width bigint,id string,is_enabled boolean,is_new boolean,name string,penny_donate string,penny_price bigint,start_date bigint,static_icon_height bigint,static_icon_url string,static_icon_width bigint,sticky_duration_seconds string,subreddit_coin_reward bigint,subreddit_id string>>",
            ):
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("award_sub_type")),
                        direct_to_binary(elem.getField("award_type")),
                        direct_to_binary(
                            elem.getField("awardings_required_to_grant_benefits")
                        ),
                        direct_to_binary(elem.getField("coin_price")),
                        direct_to_binary(elem.getField("coin_reward")),
                        direct_to_binary(elem.getField("count")),
                        direct_to_binary(elem.getField("days_of_drip_extension")),
                        direct_to_binary(elem.getField("days_of_premium")),
                        direct_to_binary(elem.getField("description")),
                        direct_to_binary(elem.getField("end_date")),
                        direct_to_binary(elem.getField("giver_coin_reward")),
                        direct_to_binary(elem.getField("icon_format")),
                        direct_to_binary(elem.getField("icon_height")),
                        direct_to_binary(elem.getField("icon_url")),
                        direct_to_binary(elem.getField("icon_width")),
                        direct_to_binary(elem.getField("id")),
                        boolean_to_binary(elem.getField("is_enabled")),
                        boolean_to_binary(elem.getField("is_new")),
                        direct_to_binary(elem.getField("name")),
                        direct_to_binary(elem.getField("penny_donate")),
                        direct_to_binary(elem.getField("penny_price")),
                        direct_to_binary(elem.getField("start_date")),
                        direct_to_binary(elem.getField("static_icon_height")),
                        direct_to_binary(elem.getField("static_icon_url")),
                        direct_to_binary(elem.getField("static_icon_width")),
                        direct_to_binary(elem.getField("sticky_duration_seconds")),
                        direct_to_binary(elem.getField("subreddit_coin_reward")),
                        direct_to_binary(elem.getField("subreddit_id")),
                    ),
                )
            case _:
                assert False, f"missed field {field_name}"


def canonicalize_discord_fields(
    data_after_norm: Column, field_specs: list[list[str]]
) -> Generator[Column, None, None]:
    for field_name, field_type in field_specs:
        match (field_name, field_type):
            case (_, "bigint" | "int" | "string" | "long"):
                yield direct_to_binary(data_after_norm.getField(field_name))
            case (_, "boolean"):
                yield boolean_to_binary(data_after_norm.getField(field_name))
            case (_, "array<string>"):
                yield array_to_binary(
                    data_after_norm.getField(field_name), direct_to_binary
                )
            case (
                _,
                ty,
            ) if ty == f"array<struct<{schema.discord_user_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name), discord_user_to_binary
                )
            case (
                "user",
                ty,
            ) if ty == f"struct<{schema.discord_user_subschema}>":
                yield discord_user_to_binary(data_after_norm.getField(field_name))
            # Channel cases
            case (
                "permission_overwrites",
                ty,
            ) if ty == f"array<struct<{schema.permission_overwrite_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("id")),
                        direct_to_binary(elem.getField("type")),
                        direct_to_binary(elem.getField("allow")),
                        direct_to_binary(elem.getField("deny")),
                    ),
                )
            case (
                "thread_metadata",
                ty,
            ) if ty == f"struct<{schema.thread_metadata_subschema}>":
                yield fn.concat(
                    boolean_to_binary(
                        data_after_norm.getField(field_name).getField("archived")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField(
                            "auto_archive_duration"
                        )
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField(
                            "archive_timestamp"
                        )
                    ),
                    boolean_to_binary(
                        data_after_norm.getField(field_name).getField("locked")
                    ),
                    boolean_to_binary(
                        data_after_norm.getField(field_name).getField("invitable")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField(
                            "create_timestamp"
                        )
                    ),
                )
            case (
                "member",
                ty,
            ) if ty == f"struct<{schema.discord_member_subschema}>":
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("id")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("user_id")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("join_timestamp")
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("flags")
                    ),
                    fn.concat(
                        discord_user_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("member")
                            .getField("user")
                        ),
                        direct_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("member")
                            .getField("nick")
                        ),
                        direct_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("member")
                            .getField("avatar")
                        ),
                        direct_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("member")
                            .getField("joined_at")
                        ),
                    ),
                )
            case (
                "available_tags",
                ty,
            ) if ty == f"array<struct<{schema.available_tags_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("id")),
                        direct_to_binary(elem.getField("name")),
                        boolean_to_binary(elem.getField("moderated")),
                        direct_to_binary(elem.getField("emoji_id")),
                        direct_to_binary(elem.getField("emoji_name")),
                    ),
                )
            case (
                "default_reaction_emoji",
                "struct<emoji_id string,emoji_name string>",
            ):
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("emoji_id"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("emoji_name"),
                    ),
                )
            # Message cases
            case (
                "author",
                ty,
            ) if ty == f"struct<{schema.discord_user_subschema}>":
                yield discord_user_to_binary(data_after_norm.getField(field_name))
            case (
                "mention_channels",
                ty,
            ) if ty == f"array<struct<{schema.mention_channel_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("id")),
                        direct_to_binary(elem.getField("guild_id")),
                        direct_to_binary(elem.getField("type")),
                        direct_to_binary(elem.getField("name")),
                    ),
                )
            case (
                "embeds",
                ty,
            ) if ty == f"array<struct<{schema.message_embed_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("title")),
                        direct_to_binary(elem.getField("type")),
                        direct_to_binary(elem.getField("description")),
                        direct_to_binary(elem.getField("url")),
                        direct_to_binary(elem.getField("timestamp")),
                        direct_to_binary(elem.getField("image").getField("url")),
                        direct_to_binary(elem.getField("video").getField("url")),
                        fn.concat(
                            direct_to_binary(elem.getField("author").getField("name")),
                            direct_to_binary(elem.getField("author").getField("url")),
                        ),
                        fn.concat(
                            direct_to_binary(
                                elem.getField("provider").getField("name")
                            ),
                            direct_to_binary(elem.getField("provider").getField("url")),
                        ),
                    ),
                )
            case (
                "reactions",
                ty,
            ) if ty == f"array<struct<{schema.reaction_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        fn.concat(
                            direct_to_binary(elem.getField("emoji").getField("id")),
                            direct_to_binary(elem.getField("emoji").getField("name")),
                        ),
                        direct_to_binary(elem.getField("count")),
                        direct_to_binary(elem.getField("burst_count")),
                        fn.concat(
                            direct_to_binary(
                                elem.getField("count_details").getField("burst")
                            ),
                            direct_to_binary(
                                elem.getField("count_details").getField("normal")
                            ),
                        ),
                        boolean_to_binary(elem.getField("me")),
                        boolean_to_binary(elem.getField("burst_me")),
                        boolean_to_binary(elem.getField("me_burst")),
                    ),
                )
            case (
                "message_reference",
                ty,
            ) if ty == f"struct<{schema.message_ref_subschema}>":
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("type"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("channel_id"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("message_id"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("guild_id"),
                    ),
                )
            case (
                "thread",
                ty,
            ) if ty == f"struct<{schema.thread_subschema}>":
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("id"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("type"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("name"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("guild_id"),
                    ),
                )
            case (
                "sticker_items",
                ty,
            ) if ty == f"array<struct<{schema.sticker_item_subschema}>>":
                yield array_to_binary(
                    data_after_norm.getField(field_name),
                    lambda elem: fn.concat(
                        direct_to_binary(elem.getField("id")),
                        direct_to_binary(elem.getField("name")),
                        direct_to_binary(elem.getField("format_type")),
                    ),
                )
            case (
                "poll",
                ty,
            ) if ty == f"struct<{schema.poll_subschema}>":
                yield fn.concat(
                    fn.concat(
                        direct_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("question")
                            .getField("text"),
                        ),
                        fn.concat(
                            direct_to_binary(
                                data_after_norm.getField(field_name)
                                .getField("question")
                                .getField("emoji")
                                .getField("id")
                            ),
                            direct_to_binary(
                                data_after_norm.getField(field_name)
                                .getField("question")
                                .getField("emoji")
                                .getField("name")
                            ),
                        ),
                    ),
                    array_to_binary(
                        data_after_norm.getField(field_name).getField("answers"),
                        lambda elem: fn.concat(
                            direct_to_binary(elem.getField("answer_id")),
                            fn.concat(
                                direct_to_binary(
                                    elem.getField("poll_media").getField("text")
                                ),
                                direct_to_binary(
                                    elem.getField("poll_media")
                                    .getField("emoji")
                                    .getField("id")
                                ),
                                direct_to_binary(
                                    elem.getField("poll_media")
                                    .getField("emoji")
                                    .getField("name")
                                ),
                            ),
                        ),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("expiry"),
                    ),
                    boolean_to_binary(
                        data_after_norm.getField(field_name).getField(
                            "allow_multiselect"
                        ),
                    ),
                    fn.concat(
                        boolean_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("results")
                            .getField("is_finalized"),
                        ),
                        array_to_binary(
                            data_after_norm.getField(field_name)
                            .getField("results")
                            .getField("answer_counts"),
                            lambda elem: fn.concat(
                                direct_to_binary(elem.getField("id")),
                                direct_to_binary(elem.getField("count")),
                                boolean_to_binary(elem.getField("me_voted")),
                            ),
                        ),
                    ),
                )
            case (
                "call",
                ty,
            ) if ty == f"struct<{schema.call_subschema}>":
                yield fn.concat(
                    array_to_binary(
                        data_after_norm.getField(field_name).getField("participants"),
                        direct_to_binary,
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField(
                            "ended_timestamp"
                        ),
                    ),
                )
            # Reaction cases
            case (
                "emoji",
                "struct<id string,name string>",
            ):
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("id"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("name"),
                    ),
                )
            # Member cases
            case (
                "avatar_decoration_data",
                "struct<asset string,sku_id string>",
            ):
                yield fn.concat(
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("asset"),
                    ),
                    direct_to_binary(
                        data_after_norm.getField(field_name).getField("sku_id"),
                    ),
                )
            case _:
                assert False, f"missed field {field_name}"


def ingest_version(
    data_after_norm: Column,
    canonicalize_field_generator: Callable[
        [Column, list[list[str]]], Generator[Column, None, None]
    ],
    field_specs: list[list[str]],
) -> Column:
    return fn.md5(
        fn.concat(*canonicalize_field_generator(data_after_norm, field_specs))
    )
