import unittest

from pyspark.sql import SparkSession
from pyspark.sql.functions import (
    col,
    md5,
    concat,
    lit,
)
from pyspark.sql.types import StringType

from trend_etl.test_utils import assert_dataframes_equal, get_kafka_df, getspark
import trend_etl.schema.fields as fields
import trend_etl.schema.schema as schema
from .ingest_versioning import (
    boolean_to_binary,
    direct_to_binary,
    array_to_binary,
    ingest_version,
    canonicalize_reddit_fields,
    canonicalize_facebook_fields,
    canonicalize_discord_fields,
)
from .parsing import (
    parse_subreddits,
    parse_reddit_submissions,
    parse_reddit_replies,
    parse_facebook_groups,
    parse_facebook_posts,
    parse_facebook_comments,
    parse_guilds,
    parse_channels,
    parse_members,
    parse_messages,
    parse_reactions,
    normalize_reddit,
)


class TestFunctions(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_direct_to_binary_null_empty_string(self) -> None:
        input_df = self.__class__.spark.createDataFrame(
            [(None,)], schema="fruit string"
        )

        got_df = input_df.withColumn(
            "ingest_version", md5(direct_to_binary(col("fruit")))
        )

        # $ echo -n -e "" | md5
        # d41d8cd98f00b204e9800998ecf8427e

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "d41d8cd98f00b204e9800998ecf8427e",
            "doc version mismatch",
        )

        input_df = self.__class__.spark.createDataFrame([("",)], schema="fruit string")

        got_df = input_df.withColumn(
            "ingest_version", md5(direct_to_binary(col("fruit")))
        )

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "d41d8cd98f00b204e9800998ecf8427e",
            "doc version mismatch",
        )

    def test_direct_to_binary(self) -> None:
        data = [("apple",)]
        input_df = self.__class__.spark.createDataFrame(data, schema="fruit string")

        got_df = input_df.withColumn(
            "ingest_version", md5(direct_to_binary(col("fruit")))
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65" | md5
        # 1f3870be274f6c49b3e31a0c6728957f

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "1f3870be274f6c49b3e31a0c6728957f",
            "doc version mismatch",
        )

    def test_boolean_to_binary(self) -> None:
        data = [(True,)]
        input_df = self.__class__.spark.createDataFrame(data, schema="flag boolean")

        got_df = input_df.withColumn(
            "ingest_version", md5(boolean_to_binary(col("flag")))
        )

        # $ echo -n -e "\x00\x00\x00\x01" | md5
        # f1450306517624a57eafbbf8ed995985

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "f1450306517624a57eafbbf8ed995985",
            "doc version mismatch",
        )

    def test_array_to_binary_array_strings(self) -> None:
        data = [(["apple"],)]
        input_df = self.__class__.spark.createDataFrame(
            data, schema="fruits array<string>"
        )

        got_df = input_df.withColumn(
            "ingest_version", md5(array_to_binary(col("fruits"), direct_to_binary))
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65" | md5
        # 1f3870be274f6c49b3e31a0c6728957f

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "1f3870be274f6c49b3e31a0c6728957f",
            "doc version mismatch",
        )

    def test_array_to_binary_array_structs(self) -> None:
        data = [
            (
                [
                    ("apple", 1, 42),  # Tuple representing the struct
                ],
            ),
        ]
        input_df = self.__class__.spark.createDataFrame(
            data,
            schema="rows array<struct<fruit string, picked_on int, number_collected int>>",
        )

        got_df = input_df.withColumn(
            "ingest_version",
            md5(
                array_to_binary(
                    col("rows"),
                    lambda elem: concat(
                        direct_to_binary(elem.getField("fruit")),
                        direct_to_binary(elem.getField("picked_on")),
                        direct_to_binary(elem.getField("number_collected")),
                    ),
                )
            ),
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65\x00\x00\x00\x01\x00\x00\x00\x2A" | md5
        # 7abbf5554a6355f2572943f1aae9a338

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "7abbf5554a6355f2572943f1aae9a338",
            "doc version mismatch",
        )

    def test_direct_to_binary_concat(self) -> None:
        data = [("apple", 1, 42)]
        input_df = self.__class__.spark.createDataFrame(
            data, schema="fruit string, picked_on int, number_collected int"
        )

        got_df = input_df.withColumn(
            "ingest_version",
            md5(
                concat(
                    direct_to_binary(col("fruit")),
                    direct_to_binary(col("picked_on")),
                    direct_to_binary(col("number_collected")),
                )
            ),
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65\x00\x00\x00\x01\x00\x00\x00\x2A" | md5
        # 7abbf5554a6355f2572943f1aae9a338

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "7abbf5554a6355f2572943f1aae9a338",
            "doc version mismatch",
        )

    def test_subreddit_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "subreddits.jsonl"
        )

        parsed_df = source_df.select(
            parse_subreddits(col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        filtered = parsed_df.where(col("json.kind") == lit("t5"))
        normalized = filtered.select(
            normalize_reddit(col("json").getField("data")).alias("norm_data")
        )
        version_added = normalized.withColumn(
            "ingest_version",
            ingest_version(
                col("norm_data"),
                canonicalize_reddit_fields,
                schema.subreddit_field_specs,
            ),
        )

        data = [
            ("AskFoo", "c1ad868b4ef32621f5f3b99af05acec9"),
            ("TheVoid", "1520e0aa7a003ef96532fe65cac91965"),
            ("HuckleberryPieAficianados", "c2ff2d5ed37f455933fb530b2c1dd7b0"),
            ("OccasionallySilly", "02fd0766cba2199e443b29316b96884b"),
            ("AskFoo", "c1ad868b4ef32621f5f3b99af05acec9"),
        ]
        got_df = version_added.selectExpr("norm_data.display_name", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_submission_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "submissions.jsonl"
        )

        parsed_df = source_df.select(
            parse_reddit_submissions(
                col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        normalized = parsed_df.select(
            normalize_reddit(col("json").getField("data")).alias("norm_data")
        )
        version_added = normalized.withColumn(
            "ingest_version",
            ingest_version(
                col("norm_data"),
                canonicalize_reddit_fields,
                schema.submissions_field_specs,
            ),
        )

        data = [
            ("t3_alvn002", "dec9395d3379a46f39a27a37abc552c1"),
            ("t3_alvn002", "eb2f956f33b78605b09065f70556d74e"),
            ("t3_alvn002", "c836dea88f3e0755fc8848aab7e0eb8f"),
            ("t3_klvg001", "4d21a879775bfca941a9b4e064c6a2c1"),
            ("t3_alvn001", "fa6bb44ec545b227fe55dd62d1c1a5e6"),
            ("t3_alvn001", "fa6bb44ec545b227fe55dd62d1c1a5e6"),
            ("t3_alvn001", "fa6bb44ec545b227fe55dd62d1c1a5e6"),
            ("t3_alvn004", "6f4d86ab40eb2af0d38530364a4cb6c6"),
            ("t3_alvn003", "581901383b81922fce183edce4276849"),
            ("t3_alvn003", "581901383b81922fce183edce4276849"),
        ]
        got_df = version_added.selectExpr("norm_data.name", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(
            data,
            schema=got_df.schema,
        )

        assert_dataframes_equal(got_df, want_df)

    def test_reply_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "replies.jsonl"
        )

        parsed_df = source_df.select(
            parse_reddit_replies(
                col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        normalized = parsed_df.select(
            normalize_reddit(col("json").getField("data")).alias("norm_data")
        )
        version_added = normalized.withColumn(
            "ingest_version",
            ingest_version(
                col("norm_data"), canonicalize_reddit_fields, schema.replies_field_specs
            ),
        )

        data = [
            ("t1_alvn101", "22f85c20ba368ff7bc8cbd94683da4d5"),
            ("t1_alvn102", "528e974339fe54781749ad12c5384d62"),
            ("t1_alvn102", "aa1c712172bedaeb57a80bc9f62ca4b4"),
            ("t1_alvn103", "71c8df932fa7075fbc0de06119306aa1"),
            ("t1_alvn104", "83af9c6da37da850ac0780175f86a2c8"),
            ("t1_alvn104", "dbd3e0513547fc7476def202ef9620f6"),
            ("t1_alvn105", "c8886f86e6291944803cb7168da52af5"),
            ("t1_alvn109", "e08bb9650b752fa8980adb393458e124"),
            ("t1_alvn100", "15e0dd716483817d5190493e9943b2a0"),
            ("t1_alvn106", "0627ee09a1719ebab87a70ef190f54fa"),
            ("t1_alvn107", "8ae977d242400667aab75eac1bc3b9a4"),
            ("t1_alvn108", "a2a22a4972c5fd8cc3d736112bd07f48"),
            ("t1_alvn108", "a2a22a4972c5fd8cc3d736112bd07f48"),
        ]
        got_df = version_added.selectExpr("norm_data.name", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(
            data,
            schema=got_df.schema,
        )

        assert_dataframes_equal(got_df, want_df)

    def test_group_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "groups.jsonl"
        )

        parsed_df = source_df.select(
            parse_facebook_groups(
                col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"), canonicalize_facebook_fields, schema.group_field_specs
            ),
        )

        data = [
            ("Brainy Conversations", "22b5aa998b1e21a23a96d979af5f0356"),
            ("Brainy Conversations", "22b5aa998b1e21a23a96d979af5f0356"),
            ("Chris Test Group", "75b99007d2f34cfb863887d6c5cc45f1"),
            ("Chris Test Group", "75b99007d2f34cfb863887d6c5cc45f1"),
        ]
        got_df = version_added.selectExpr("json.name", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_posts_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "posts.jsonl"
        )

        parsed_df = source_df.select(
            parse_facebook_posts(
                col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"), canonicalize_facebook_fields, schema.posts_field_specs
            ),
        )

        data = [
            ("100_1", "ea58eed04451165798b65e4ba7950c19"),
            ("100_2", "32ec2588f0613513fc9b773c79043607"),
            ("100_3", "33d583cd0510a36dfa4c7e967c89cd3c"),
            ("100_4", "1bad4e25bd170d638cb1a21e7655ca24"),
            ("100_6", "f8f366249698f73afe2288172034a18f"),
            ("100_6", "f8f366249698f73afe2288172034a18f"),
            ("200_1", "e9c1d4cbb94e3e045c0b1d5dd0c2e6be"),
        ]
        got_df = version_added.selectExpr("json.id", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_comments_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "comments.jsonl"
        )

        parsed_df = source_df.select(
            parse_facebook_comments(
                col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"), canonicalize_facebook_fields, schema.comments_field_specs
            ),
        )

        data = [
            ("1", "60ec22591f00f72f5153427b9ed34af9"),
            ("2", "0d1502b10cefd11c149745f46c9fdeaf"),
            ("3", "94abe557bf5e373968d57dd50bdf7f94"),
            ("4", "c39960d5eeafd77d4358266dfec233b7"),
            ("5", "607f04f1118273b501cfd1f10d2ff45e"),
            ("6", "a9bf059249e7be5fa460b846170650ba"),
            ("7", "84898172174d72515b41552a4a208dc9"),
            ("8", "f0f6d1b01b74414f3a3a41409110b367"),
            ("9", "fa9222878b581e916d0caca64c0bb27b"),
            ("14", "f10232ff6efa15eb8aef78c8e08d1ab5"),
            ("14", "f10232ff6efa15eb8aef78c8e08d1ab5"),
            ("10", "bf853e9619972addd4bc49bd4ff69c3f"),
            ("11", "6d2ebf8a749fe445d1dae0f31e81c0f6"),
            ("12", "6680370b47705d6d007ce6550c11b197"),
            ("13", "9f836170d0bf1f139060f5aae45389c0"),
        ]
        got_df = version_added.selectExpr("json.id", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_guild_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "guilds.jsonl"
        )

        parsed_df = source_df.select(
            parse_guilds(col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"),
                canonicalize_discord_fields,
                schema.discord_guild_field_specs,
            ),
        )

        data = [
            ("TREND Community", "ec11a32c2463d0c542cdb40d046877ed"),
            ("TREND Community", "ec11a32c2463d0c542cdb40d046877ed"),
        ]
        got_df = version_added.selectExpr("json.name", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_channel_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "channels.jsonl"
        )

        parsed_df = source_df.select(
            parse_channels(col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"),
                canonicalize_discord_fields,
                schema.discord_channel_field_specs,
            ),
        )

        data = [
            ("1205270233926074520", "bb238adce58525a6a7c05f93bd6a5802"),
            ("1205270233926074523", "bf157546d20e4a91c816d9dbee1e1660"),
            ("1205270233926074525", "fca92847cbe63689d1f64cb7af1aaa4b"),
            ("1205270233926074526", "c511850735b12960fb15893870e5c99c"),
            ("1205270233926074527", "127b015c22a4ce2cf372446115e4c8c1"),
            ("1205270234307760128", "8e83ec378d42a47b9347d0f4a14e3e5a"),
            ("1205270234307760130", "3921b76e6ee46f9f4f1a57d14b55f2b4"),
            ("1205270234307760132", "f6684659f5b77f5ea28829ea08fdb7fd"),
            ("1205270234307760134", "2f6e9e18b8db24db83be0c94e1e39001"),
            ("1205270234307760135", "adedfdff8bd1fa61dba195b51e9dba0b"),
            ("1205270234307760136", "6eacd733d99256922215a8b59521bf71"),
            ("1215765557513556170", "bb8dbc6a9660b7442f08f4cfd14ae327"),
            ("1215765928013201488", "d9ecd5a4ffa95dd6bfbf3b5e08164e91"),
            ("1219682490059329726", "8fc17c9dae20bc84f883048995a13eef"),
            ("1215775515412004904", "05b4cbc9d678a69a30faa689c7e358c1"),
            ("1225217589853028483", "9793adeefd89992757bf0151952e1908"),
            ("1225532197700898856", "5440b7f4c92390f0a544a50a7092acbb"),
            ("1225534550113910995", "6b7cadff918cd35c9d61d5bf502847c2"),
            ("1225883121459925033", "93d1903d7d08a59f6c9d80331c632722"),
            ("1227702727635763214", "67425c81238e5c0daa6798651c220a4e"),
            ("1227705593217683536", "1b5fd0125d091705a0f2996d04ff211b"),
            ("1227706023750537347", "6dbbd80ac6247d7a1d7e0b6c5858eb70"),
            ("1227706310825476167", "93ca172665c59968e19785b04890ae94"),
            ("1227707869860728862", "867a327cf727243d069ec1a9669ed7d3"),
            ("1228426021574021200", "f62933361ab95f22602e49ef93185ab1"),
            ("1228426235215089696", "98c7d89b41567a0760d74cba88368fe2"),
            ("1235969369540923455", "eed55f71a625cc02b9b67c8e30d97877"),
            ("1215766548124532757", "3c01f8f3c35e6dca7e52e531332d3f89"),
        ]
        got_df = version_added.selectExpr("json.id", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_message_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "messages.jsonl"
        )

        parsed_df = source_df.select(
            parse_messages(col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"),
                canonicalize_discord_fields,
                schema.discord_message_field_specs,
            ),
        )

        data = [
            ("1206672250913296444", "b01f4c63e86d3522af8783bed63ef71c"),
            ("1206672232999551008", "2786b71c92d0584c5acbda28f1eb1341"),
            ("1206672217161596940", "9c55773e5c969fbb6045f34b3e3b9b23"),
            ("1206672156872810537", "1446d65c646e74cb1649512831a2aa4c"),
            ("1245878725111709766", "be49712d3a3a9667eb1dd5352771e419"),
            ("1245515631714107542", "2fb4de1fadf72a260cde9aea7aa88127"),
            ("1230183182523957308", "3e3d5eb0f346ca689815153726dfe454"),
            ("1227992918757740696", "6a66b9c48b8253aaa2f933d2871cc52a"),
            ("1227953984405372989", "df2964d67b415e71b87446827ca8ae2e"),
            ("1227953850602754070", "5906e1956b495e44dac41a3966c2a7e7"),
            ("1227953764930031657", "aece4d4fbe92e05666a165ce73692fd7"),
            ("1227953480921124896", "7c6c05499903b3e2814df158a9781572"),
            ("1227953366957559838", "01241fd4a0fc5493e297159b53f9fac9"),
            ("1227952939541467186", "ae91ee81c9a74e1a6fda02c10354cadb"),
            ("1227952876542889994", "5a34865ef4f4d9ac7d5a79b087cc0617"),
            ("1227952324375351366", "9fcb37d6d985b9cf5084942c9ef98403"),
            ("1227737919230443541", "9761ed59d6ca66a3ba07e50ef44f5a48"),
            ("1227713782831386805", "0e173841ee26651b9beab6dfc5857ab4"),
            ("1227708690979618817", "7c5ade6fc47a447a2b77f15f856677a8"),
            ("1227327559314706583", "ded9ee12988c3b9ab4ba58342de30ead"),
            ("1227323501082579116", "8bc4c28d7bbcb189921de06834ebb545"),
            ("1226915463804489919", "4421b19d82c14c03f103d4ec44efdef6"),
            ("1226675679790370887", "e03ecf96006fae44e8c354cca0eddac3"),
            ("1226671358839099432", "f92d49fb2c3e598211f18402e0925018"),
            ("1226659365994627073", "d101e3e9ce5b95aa76732b7eba9d8dbc"),
            ("1226630757548560458", "85cd88c7df06c9a39bb3fcfc59cfd3a1"),
            ("1225851496239726614", "79d973b83d4896434cae3aacb7c3487e"),
            ("1225851493744250890", "788a7c50b2bb15e3e54010ef546cec42"),
            ("1225108879520825464", "20bf50735e1bff882d4be1a9c474dc6e"),
            ("1225108875670327357", "f84b8c717b937daa7095f740e8e038d6"),
            ("1215309179791605841", "d4458e15ff8b1b1dbd77c096c1cf2385"),
            ("1214945291673800794", "1267dc591f37f5a130986847f92c26d6"),
            ("1214634327678259261", "b5e281ab58929132ab317ab2f18d7acd"),
            ("1214591509932933120", "10a669c701cee645858f57e51cfff978"),
            ("1214347532822388777", "5994e1ad2fd2eb0bdb704f4bb1c262ae"),
            ("1209954849785909348", "a27d34fe7678a93af07350f763b53944"),
            ("1207713091945496637", "2c080b0fd281f064e1019a4906ede826"),
            ("1207713016586436643", "a39aa5dbffaacc39234333154249502a"),
            ("1207419490137153636", "b95156f0edf5e385faa8f8dfec56758c"),
            ("1207415009181966377", "b0496f7ab3623e1863448033ca34a9f5"),
            ("1207414936029106186", "deb5257e86875e535c9487ab0906309c"),
            ("1206981167560134696", "c14e4f00568a55535d292605d0ebb659"),
            ("1206699020722896936", "62cacdac227e7eff88da45e7e5f09297"),
            ("1206689393130348544", "632e6b818045201085f4b7b41e8714ed"),
            ("1206688988337934336", "07c9bfeb9d3e496aa26e444f1dd89572"),
            ("1206688843697356810", "91aa61ed7dc71419176ebc2e85338f43"),
            ("1206688579690963034", "6e1694f62fce6944df0659777c4e85b6"),
            ("1206677185239195808", "0aa06071d4ae885de4bd3d8c6fac5995"),
            ("1206677112396709909", "41b078c3a121a910b66f0da9b938d7b3"),
            ("1205271242710716536", "9da7ed7e806df74aa13caeb558ef1c4e"),
            ("1225882645984972831", "3c10d2348b5179be657562d8ebe6350e"),
            ("1225882611038031952", "5db4306497efd247d342e6d84f523068"),
            ("1225882570416197683", "da0f45b3cbd147c9cad7546b5e13f293"),
            ("1225882493819818136", "5d23ee100b3d858d32e6760460f81fba"),
            ("1225882427784691753", "fb66ff8beaf9e99ba74cc85c7d83e465"),
            ("1225882368150212740", "915b8b64755181fe1c3f0eb1b4838aa3"),
            ("1225882319987146783", "84a887375058efbf01cfd49b82200230"),
            ("1220066846095446036", "340284adccf3cf9310ae6f7bffa78e75"),
            ("1219682490059329726", "f6ac30466eb30c360c453b9642d0e838"),
            ("1215775515412004904", "9a87354217e553b888d425826e75e5c1"),
            ("1225346878648881183", "a4cd3254e908833c3d113e00b43745f4"),
            ("1227352559728066581", "62221f28bf15cc0b30735e95a5c4ad72"),
            ("1227294371376795678", "167ea41558bc46f302e6542753e04923"),
            ("1226915927455563848", "ab4245b1904e22ffd420b93c78a78558"),
            ("1226351219149111306", "d0f62de875eb923aa2d773bb29d8a9a5"),
            ("1225895035703132231", "9395b1da0d654d3fc2415ebc937e07bf"),
            ("1225534550113910995", "b441907e0ddd039039755945bf5352f4"),
            ("1228448533472870612", "94645f03d2ef21e9626ec485879d6563"),
            ("1230234115622699109", "8ec92d3ff26d979d2ca9f96da736faff"),
            ("1230234106907070586", "19c69713e89db92fbdb4f8374cfa9234"),
            ("1229518402239070308", "d6994ff50f39645accc8901063171883"),
            ("1229518400896897034", "c2ee34feb5366e9e2ed44b3970a2cd7e"),
            ("1237052919258550348", "f22a0cdaac74ab64799a77d889a41c47"),
            ("1235969369540923455", "8965a8ccef0dca622e0920adde8a8cb0"),
            ("1237052920390881351", "66e46fc0e8ea9ec2673f47818250e7db"),
            ("1237052918717349898", "9a7124a30af9401a0d1ba9897564e243"),
            ("1216758946036191322", "d1a48ff059aa0a9ae1aab61f01d1c98d"),
            ("1215774261550252053", "4e71b28c763803281502535b26c67ed3"),
            ("1215766548124532757", "025d3dd312469af26bae09495ca61812"),
            ("1250462460184498237", "dad37522349416768cde39c8a9ce9754"),
            ("1271205123892449606", "cabf3270d111d368331902e559bbb4f1"),
            ("1227713782831386805", "0e173841ee26651b9beab6dfc5857ab4"),
        ]
        got_df = version_added.selectExpr("json.id", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_reaction_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "reactions.jsonl"
        )

        parsed_df = source_df.select(
            parse_reactions(col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"),
                canonicalize_discord_fields,
                schema.discord_reaction_field_specs,
            ),
        )

        data = [
            ("1206672250913296444", "ea28b455691a9d7e48ac12b02ad294f1"),
            ("1206672250913296444", "95a57dbdce19e4dffe4e70d45bd996aa"),
            ("1206672250913296444", "d9def5f4c2c6fe2c1755814f53994eaf"),
            ("1206672250913296444", "efdf1cda33d9f7e5eb04aeb9d3e81073"),
            ("1206672250913296444", "148dd603fc7c45fdfada75905f103627"),
            ("1206672250913296444", "c0239669fb116f4b1d0dee801aac7724"),
            ("1206672250913296444", "79582c261f88f0b10152373c9e9078eb"),
            ("1206672250913296444", "c50c013341de401a932c0774795f64d3"),
            ("1206672250913296444", "d24890f11af8cfa4cf59bf726c0ea120"),
            ("1206672250913296444", "7de3807c6996d06f3c8fa4824f0c9eae"),
            ("1206672250913296444", "e4a3cb8d5efff914c0d8a96602217625"),
            ("1206672250913296444", "58fa2526dcee2fdc7267195625031fe7"),
            ("1206672250913296444", "d972ce88456a7ff012ef8265e3258e90"),
            ("1206672250913296444", "33e57f9f25378b54284f513b5dd91f5b"),
            ("1206672232999551008", "8255ee5da34bb97953e0dae725a64d4f"),
            ("1206672232999551008", "55d199f56096cc02132e5e7399a2c36e"),
            ("1206672232999551008", "532ae946883720157490c7c46ff3fcfa"),
            ("1206672232999551008", "dd55dd5704fcebc70291e7b23d15f900"),
            ("1206672232999551008", "7b183b30745e9bc94fded0fcb3a5e226"),
            ("1206672232999551008", "3cccd5868bcb288e3f487513d9cd2857"),
            ("1206672232999551008", "2aade6a33b4b4b8ba5e6f5451cbde60a"),
            ("1206672217161596940", "14a00c27a9b89477d82c52d9e9744f1c"),
            ("1206672217161596940", "72ec8102c3fc5e8daefcbfbe0d53dcea"),
            ("1206672217161596940", "8d6a05c325f71728dfd8444908e115dd"),
            ("1206672217161596940", "86e25c31d2018ab2bd455a307d618bd9"),
            ("1206672217161596940", "5ee829ccaaa9900c585009960eb53921"),
            ("1206672217161596940", "cd8c13eb18a59acb2b2bb509ce6fe9f6"),
            ("1206672217161596940", "d47d257003e7989657ca18a4769a3655"),
            ("1206672217161596940", "8a6a1c4cd9cee2795b6317ab9254dedb"),
            ("1206672217161596940", "cf02654922ffa9fef7a02eb2473df82f"),
            ("1206672217161596940", "73bf02937904866fbd5fb5d0a3686ac5"),
            ("1230183182523957308", "7618eff9c35993ae9659118ede2126b7"),
            ("1227953984405372989", "41cad4825aaf99c1e9e4289597ff5755"),
            ("1227953850602754070", "acb3932e384ab07475b95030cfc2313b"),
            ("1227952939541467186", "72cbace099928bb49f3f7a1e3bfe540c"),
            ("1227952324375351366", "66c7d8b54c0099ce8832f872285b71e5"),
            ("1227952324375351366", "fba73c116be2b8db4fbeb877ce23b1b8"),
            ("1209954849785909348", "27a017c55fe9c314a61b4e7ed798d60f"),
            ("1209954849785909348", "edb5058acd557ed19680a5af3512861c"),
            ("1219682490059329726", "1acc054979806049e979572257871ce1"),
            ("1215775515412004904", "69b0c28730b58fd8c599a1977317e2e1"),
            ("1215775515412004904", "fddd30d7442581173bcb1c27ebd11cf6"),
            ("1227294371376795678", "1106bd29b60293f7dec324723d65f776"),
            ("1227294371376795678", "bddb16ff9df8a11889d1378402bff3a2"),
            ("1226915927455563848", "68a7cadb4e4880402a79778e783cab3b"),
            ("1226351219149111306", "a954fd2c25307fa3e74a570edbf0c735"),
            ("1230234115622699109", "f83282fe2b89446a02c5c402f81f4607"),
            ("1229518402239070308", "cead45c9850eeca0fb93ad1a85359f89"),
            ("1229518402239070308", "1b978684f7c759c2a3ae90f0b73b408a"),
            ("1237052919258550348", "2c0504d5442f25f7974d0417ac21b146"),
            ("1237052919258550348", "d744119559c5a2e197c805684c4d175d"),
            ("1235969369540923455", "230de03aa4a8ff1f9d8f1c4ee79b0c0a"),
            ("1235969369540923455", "a55a215af1da3d91f5c89e84430d4271"),
            ("1235969369540923455", "06610f49ee4c35d3d6d513284ea0eb75"),
            ("1216758946036191322", "2878e249687d1722f4858898f786769d"),
            ("1215766548124532757", "d96744dde991214eff5794e9d74c230b"),
            ("1215766548124532757", "d5323be84adbcfedeaee49320af8226d"),
            ("1215766548124532757", "d7205ec2397dcfd22588e16ae48b1909"),
            ("1215766548124532757", "dcb0a9a4453a5385af57d6f1772af17d"),
        ]
        got_df = version_added.selectExpr("json.message_id", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)

    def test_member_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "members.jsonl"
        )

        parsed_df = source_df.select(
            parse_members(col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        version_added = parsed_df.withColumn(
            "ingest_version",
            ingest_version(
                col("json"),
                canonicalize_discord_fields,
                schema.discord_member_field_specs,
            ),
        )

        data = [
            ("707234442820845568", "61e53d2211a3f12b21d3a8d0b2ce97e2"),
            ("741682757486510081", "d883ffac5add495fe2574c63956d0004"),
            ("829877816228970556", "baa20afc34f9282ba3e96f4aeeb03676"),
            ("943888971900006450", "dc57acc4b317b5819270d51cb6a933e8"),
            ("980068258084450374", "f00f975514058a634a4f0e504c5722de"),
            ("1114209896892739644", "23e9a62e3ba7ba49e2d68b3df8412c24"),
            ("1178799745133711380", "14c87f0c9c4429f8613c795be1b0e959"),
            ("1179456832159354930", "3b8febf2a69fda2b98d0bc9e13d42486"),
            ("1192554388649164863", "cf5826daf72828a65a0741f2db2a5a39"),
            ("1197548825750024232", "8c64373806a781463cb9c713639ab303"),
            ("1205240581198971001", "5a2d4301cc2451ef3bb237f21359846b"),
            ("1214634228432642098", "b2bf4e4fcd6841a2a191b73f3894937d"),
            ("1214964353275666504", "a6dfc37ec2e8afaca0447d89cffc0b4a"),
            ("1225031472432418837", "2370013f55beb2fbeebe457bfd0ed365"),
        ]
        got_df = version_added.selectExpr("json.user.id", "ingest_version")
        want_df = self.__class__.spark.createDataFrame(data, schema=got_df.schema)

        assert_dataframes_equal(got_df, want_df)
