from typing import Generator

import boto3
from mypy_boto3_s3.paginator import ListObjectsPaginator
from mypy_boto3_s3.type_defs import ListObjectsOutputTypeDef

from pyspark.sql import Column, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.types import (
    TimestampType,
    IntegerType,
)
from pyspark.sql.window import Window

import trend_etl.schema.fields as fields
from trend_etl.schema.types import S3_Bucket


# arguments:
#   field: name of the field which contains a MongoDB job id string
def jobid_2_timestamp(field: str) -> Column:
    return (
        fn.conv(fn.col(field).substr(0, 8), 16, 10)
        .cast(IntegerType())
        .cast(TimestampType())
    )


# arguments
#   input_df: a dataframe with the reactions_* columns
# return: dataframe with reactions column as a struct
def reactions_struct(input_df: DataFrame) -> DataFrame:
    reactions = fn.struct(
        fn.col("reactions_angry.summary.total_count").alias("angry"),
        fn.col("reactions_care.summary.total_count").alias("care"),
        fn.col("reactions_haha.summary.total_count").alias("haha"),
        fn.col("reactions_like.summary.total_count").alias("like"),
        fn.col("reactions_love.summary.total_count").alias("love"),
        fn.col("reactions_none.summary.total_count").alias("none"),
        fn.col("reactions_pride.summary.total_count").alias("pride"),
        fn.col("reactions_sad.summary.total_count").alias("sad"),
        fn.col("reactions_thankful.summary.total_count").alias("thankful"),
        fn.col("reactions_wow.summary.total_count").alias("wow"),
    )

    return input_df.withColumn(
        "reactions",
        reactions.alias(fields.reactions_struct_field),
    ).drop(
        "reactions_angry",
        "reactions_care",
        "reactions_haha",
        "reactions_like",
        "reactions_love",
        "reactions_none",
        "reactions_pride",
        "reactions_sad",
        "reactions_thankful",
        "reactions_wow",
    )


def get_created_day(field: Column) -> Column:
    return fn.coalesce(fn.date_format(field, "yyyy-MM-dd"), fn.lit("none"))


def get_collected_day(field: Column) -> Column:
    return fn.date_format(field, "yyyy-MM-dd")


def s3_list_folders_prefix(
    bucket: S3_Bucket,
    prefix: str,
) -> Generator[str, None, None]:
    client = boto3.client("s3")
    paginator: ListObjectsPaginator = client.get_paginator("list_objects")

    for item in paginator.paginate(Bucket=bucket, Delimiter="/", Prefix=prefix).search(
        "CommonPrefixes"
    ):
        yield item.get("Prefix")


def s3_list_objects(
    bucket: S3_Bucket,
) -> Generator[ListObjectsOutputTypeDef, None, None]:
    client = boto3.client("s3")
    paginator: ListObjectsPaginator = client.get_paginator("list_objects")

    for item in paginator.paginate(Bucket=bucket):
        yield item


def get_edited_flag_dataframe(
    input_df: DataFrame, group_key: str, compare_key: str
) -> DataFrame:
    window_spec = Window.partitionBy(group_key)

    edited_flag_df = input_df.withColumn(
        "_edited_flag",
        fn.when(
            fn.size(fn.collect_set(compare_key).over(window_spec)) > 1, fn.lit(True)
        ).otherwise(fn.lit(False)),
    )

    return edited_flag_df
